# Copyright (c) towncrier contributors, 2025
# See LICENSE for details.

import os.path
import shutil
import unittest

from pathlib import Path
from subprocess import check_call

from click.testing import CliRunner
from twisted.trial.unittest import TestCase

from towncrier import _hg, _vcs

from .helpers import setup_simple_project, write


hg_available = shutil.which("hg") is not None


def create_project(
    pyproject_path="pyproject.toml", main_branch="main", extra_config=""
):
    setup_simple_project(pyproject_path=pyproject_path, extra_config=extra_config)
    Path("foo/newsfragments/123.feature").write_text("Adds levitation")
    initial_commit(branch=main_branch)
    check_call(["hg", "branch", "otherbranch"])


def initial_commit(branch="main"):
    """
    Create a mercurial repo, configure it and make an initial commit

    There must be uncommitted changes
    """
    check_call(["hg", "init", "."])
    check_call(["hg", "branch", branch])
    commit("Initial Commit")


def commit(message):
    """Commit the repo in the current working directory

    There must be uncommitted changes
    """
    check_call(["hg", "addremove", "."])
    check_call(["hg", "commit", "--user", "Example <test@example.com>", "-m", message])


@unittest.skipUnless(hg_available, "requires 'mercurial' to be installed")
class TestHg(TestCase):
    def test_get_default_compare_branch(self):
        """
        Test the 'get_default_compare_branch' behavior.
        """
        assert _hg.get_default_compare_branch(["main", "default"]) == "default"
        assert _hg.get_default_compare_branch(["main", "a_topic"]) is None

    def test_empty_remove(self):
        """
        If remove_files gets an empty list, it returns gracefully.
        """
        _hg.remove_files([])

    def test_complete_scenario(self):
        """
        Tests all the _hg functions that interact with an actual repository.

        Setting up a project is a little slow, hence the grouping of all the
        tests in one.
        """
        runner = CliRunner()
        with runner.isolated_filesystem():
            create_project("pyproject.toml")

            # make sure _vcs._get_mod properly detects a mercurial repo
            self.assertEqual(_hg, _vcs._get_mod("."))

            write("changes/000.misc.rst", "some change")

            _hg.stage_newsfile(".", "changes/000.misc.rst")

            commit("commit 1")

            branches = sorted(_hg.get_remote_branches("."))
            self.assertEqual(["main", "otherbranch"], branches)

            self.assertEqual(
                [os.path.join("changes", "000.misc.rst")],
                _hg.list_changed_files_compared_to_branch(".", "main", False),
            )

            write("changes/001.misc.rst", "some change")
            _hg.remove_files(["changes/000.misc.rst", "changes/001.misc.rst"])



# Copyright (c) towncrier contributors, 2025
# See LICENSE for details.

from __future__ import annotations

import os

from collections.abc import Container
from subprocess import STDOUT, call, check_output


def get_default_compare_branch(branches: Container[str]) -> str | None:
    if "default" in branches:
        return "default"
    return None


def remove_files(fragment_filenames: list[str]) -> None:
    if not fragment_filenames:
        return

    # Filter out files that are unknown to mercurial
    hg_fragments = (
        check_output(["hg", "files"] + fragment_filenames, encoding="utf-8")
        .strip()
        .split("\n")
    )

    hg_fragments = [os.path.abspath(f) for f in hg_fragments if os.path.isfile(f)]
    fragment_filenames = [
        os.path.abspath(f) for f in fragment_filenames if os.path.isfile(f)
    ]
    call(["hg", "rm", "--force"] + hg_fragments, encoding="utf-8")
    unknown_fragments = set(fragment_filenames) - set(hg_fragments)
    for unknown_fragment in unknown_fragments:
        os.remove(unknown_fragment)


def stage_newsfile(directory: str, filename: str) -> None:
    call(["hg", "add", os.path.join(directory, filename)])


def get_remote_branches(base_directory: str) -> list[str]:
    branches = check_output(
        ["hg", "branches", "--template", "{branch}\n"],
        cwd=base_directory,
        encoding="utf-8",
    ).splitlines()

    return branches


def list_changed_files_compared_to_branch(
    base_directory: str, compare_with: str, include_staged: bool
) -> list[str]:
    output = check_output(
        ["hg", "diff", "--stat", "-r", compare_with],
        cwd=base_directory,
        encoding="utf-8",
        stderr=STDOUT,
    ).splitlines()

    return [line.split("|")[0].strip() for line in output if "|" in line]



# Copyright (c) Amber Brown, 2015
# See LICENSE for details.

import os

from textwrap import dedent

from click.testing import CliRunner
from twisted.trial.unittest import TestCase

from .._settings import ConfigError, load_config
from .._shell import cli
from .helpers import with_isolated_runner, write


class TomlSettingsTests(TestCase):
    def mktemp_project(
        self, *, pyproject_toml: str = "", towncrier_toml: str = ""
    ) -> str:
        """
        Create a temporary directory with a pyproject.toml file in it.
        """
        project_dir = self.mktemp()
        os.makedirs(project_dir)

        if pyproject_toml:
            write(
                os.path.join(project_dir, "pyproject.toml"),
                pyproject_toml,
                dedent=True,
            )

        if towncrier_toml:
            write(
                os.path.join(project_dir, "towncrier.toml"),
                towncrier_toml,
                dedent=True,
            )

        return project_dir

    def test_base(self):
        """
        Test a "base config".
        """
        project_dir = self.mktemp_project(
            pyproject_toml="""
                [tool.towncrier]
                package = "foobar"
                orphan_prefix = "~"
            """
        )

        config = load_config(project_dir)
        self.assertEqual(config.package, "foobar")
        self.assertEqual(config.package_dir, ".")
        self.assertEqual(config.filename, "NEWS.rst")
        self.assertEqual(config.underlines, ("=", "-", "~"))
        self.assertEqual(config.orphan_prefix, "~")

    def test_markdown(self):
        """
        If the filename references an .md file and the builtin template doesn't have an
        extension, add .md rather than .rst.
        """
        project_dir = self.mktemp_project(
            pyproject_toml="""
                [tool.towncrier]
                package = "foobar"
                filename = "NEWS.md"
            """
        )

        config = load_config(project_dir)

        self.assertEqual(config.filename, "NEWS.md")

        self.assertEqual(config.template, ("towncrier.templates", "default.md"))

    def test_explicit_template_extension(self):
        """
        If the filename references an .md file and the builtin template has an
        extension, don't change it.
        """
        project_dir = self.mktemp_project(
            pyproject_toml="""
                [tool.towncrier]
                package = "foobar"
                filename = "NEWS.md"
                template = "towncrier:default.rst"
            """
        )

        config = load_config(project_dir)

        self.assertEqual(config.filename, "NEWS.md")
        self.assertEqual(config.template, ("towncrier.templates", "default.rst"))

    def test_template_extended(self):
        """
        The template can be any package and resource, and although we look for a
        resource's 'templates' package, it could also be in the specified resource
        directly.
        """
        project_dir = self.mktemp_project(
            pyproject_toml="""
                [tool.towncrier]
                package = "foobar"
                template = "towncrier.templates:default.rst"
            """
        )

        config = load_config(project_dir)

        self.assertEqual(config.template, ("towncrier.templates", "default.rst"))

    def test_incorrect_single_file(self):
        """
        single_file must be a bool.
        """
        project_dir = self.mktemp_project(
            pyproject_toml="""
                [tool.towncrier]
                single_file = "a"
            """
        )

        with self.assertRaises(ConfigError) as e:
            load_config(project_dir)

        self.assertEqual(e.exception.failing_option, "single_file")

    def test_incorrect_all_bullets(self):
        """
        all_bullets must be a bool.
        """
        project_dir = self.mktemp_project(
            pyproject_toml="""
                [tool.towncrier]
                all_bullets = "a"
            """
        )

        with self.assertRaises(ConfigError) as e:
            load_config(project_dir)

        self.assertEqual(e.exception.failing_option, "all_bullets")

    def test_mistype_singlefile(self):
        """
        singlefile is not accepted, single_file is.
        """
        project_dir = self.mktemp_project(
            pyproject_toml="""
                [tool.towncrier]
                singlefile = "a"
            """
        )

        with self.assertRaises(ConfigError) as e:
            load_config(project_dir)

        self.assertEqual(e.exception.failing_option, "singlefile")

    def test_towncrier_toml_preferred(self):
        """
        Towncrier prefers the towncrier.toml for autodetect over pyproject.toml.
        """
        project_dir = self.mktemp_project(
            towncrier_toml="""
                [tool.towncrier]
                package = "a"
            """,
            pyproject_toml="""
                [tool.towncrier]
                package = "b"
            """,
        )

        config = load_config(project_dir)
        self.assertEqual(config.package, "a")

    def test_pyproject_only_pyproject_toml(self):
        """
        Towncrier will fallback to the [project.name] value in pyproject.toml.

        This tests asserts that the minimal configuration is to do *nothing*
        when using a pyproject.toml file.
        """
        project_dir = self.mktemp_project(
            pyproject_toml="""
                [project]
                name = "a"
            """,
        )

        config = load_config(project_dir)
        self.assertEqual(config.package, "a")
        self.assertEqual(config.name, "a")

    def test_pyproject_assert_fallback(self):
        """
        This test is an extensive test of the fallback scenarios
        for the `package` and `name` keys in the towncrier section.

        It will fallback to pyproject.toml:name in any case.
        And as such it checks the various fallback mechanisms
        if the fields are not present in the towncrier.toml, nor
        in the pyproject.toml files.

        This both tests when things are *only* in the pyproject.toml
        and default usage of the data in the towncrier.toml file.
        """
        pyproject_toml = dedent(
            """
        [project]
        name = "foo"
        [tool.towncrier]
        """
        )
        towncrier_toml = dedent(
            """
        [tool.towncrier]
        """
        )
        tests = [
            "",
            "name = '{name}'",
            "package = '{package}'",
            "name = '{name}'",
            "package = '{package}'",
        ]

        def factory(name, package):
            def func(test):
                return dedent(test).format(name=name, package=package)

            return func

        for pp_fields in map(factory(name="a", package="b"), tests):
            pp_toml = pyproject_toml + pp_fields
            for tc_fields in map(factory(name="c", package="d"), tests):
                tc_toml = towncrier_toml + tc_fields

                # Create the temporary project
                project_dir = self.mktemp_project(
                    pyproject_toml=pp_toml,
                    towncrier_toml=tc_toml,
                )

                # Read the configuration file.
                config = load_config(project_dir)

                # Now the values depend on where the fallback
                # is.
                # If something is in towncrier.toml, it will be preferred
                # name fallsback to package
                if "package" in tc_fields:
                    package = "d"
                else:
                    package = "foo"
                self.assertEqual(config.package, package)

                if "name" in tc_fields:
                    self.assertEqual(config.name, "c")
                else:
                    # fall-back to package name
                    self.assertEqual(config.name, package)

    @with_isolated_runner
    def test_load_no_config(self, runner: CliRunner):
        """
        Calling the root CLI without an existing configuration file in the base directory,
        will exit with code 1 and an informative message is sent to standard output.
        """
        temp = self.mktemp()
        os.makedirs(temp)

        result = runner.invoke(cli, ("--dir", temp))

        self.assertEqual(
            result.output,
            f"No configuration file found.\nLooked back from: {os.path.abspath(temp)}\n",
        )
        self.assertEqual(result.exit_code, 1)

    @with_isolated_runner
    def test_load_explicit_missing_config(self, runner: CliRunner):
        """
        Calling the CLI with an incorrect explicit configuration file will exit with
        code 1 and an informative message is sent to standard output.
        """
        config = "not-there.toml"
        result = runner.invoke(cli, ("--config", config))

        self.assertEqual(result.exit_code, 1)
        self.assertEqual(
            result.output,
            f"Configuration file '{os.path.abspath(config)}' not found.\n",
        )

    def test_missing_template(self):
        """
        Towncrier will raise an exception saying when it can't find a template.
        """
        project_dir = self.mktemp_project(
            towncrier_toml="""
                [tool.towncrier]
                template = "foo.rst"
            """
        )

        with self.assertRaises(ConfigError) as e:
            load_config(project_dir)

        self.assertEqual(
            str(e.exception),
            "The template file '{}' does not exist.".format(
                os.path.normpath(os.path.join(project_dir, "foo.rst")),
            ),
        )

    def test_missing_template_in_towncrier(self):
        """
        Towncrier will raise an exception saying when it can't find a template
        from the Towncrier templates.
        """
        project_dir = self.mktemp_project(
            towncrier_toml="""
                [tool.towncrier]
                template = "towncrier:foo"
            """
        )

        with self.assertRaises(ConfigError) as e:
            load_config(project_dir)

        self.assertEqual(
            str(e.exception), "'towncrier' does not have a template named 'foo.rst'."
        )

    def test_custom_types_as_tables_array_deprecated(self):
        """
        Custom fragment categories can be defined inside
        the toml config file using an array of tables
        (a table name in double brackets).

        This functionality is considered deprecated, but we continue
        to support it to keep backward compatibility.
        """
        project_dir = self.mktemp_project(
            pyproject_toml="""
                [tool.towncrier]
                package = "foobar"
                [[tool.towncrier.type]]
                directory="foo"
                name="Foo"
                showcontent=false

                [[tool.towncrier.type]]
                directory="spam"
                name="Spam"
                showcontent=true

                [[tool.towncrier.type]]
                directory="auto"
                name="Automatic"
                showcontent=true
                check=false
            """
        )
        config = load_config(project_dir)
        expected = [
            (
                "foo",
                {
                    "name": "Foo",
                    "showcontent": False,
                    "check": True,
                },
            ),
            (
                "spam",
                {
                    "name": "Spam",
                    "showcontent": True,
                    "check": True,
                },
            ),
            (
                "auto",
                {
                    "name": "Automatic",
                    "showcontent": True,
                    "check": False,
                },
            ),
        ]
        expected = dict(expected)
        actual = config.types
        self.assertDictEqual(expected, actual)

    def test_custom_types_as_tables(self):
        """
        Custom fragment categories can be defined inside
        the toml config file using tables.
        """
        project_dir = self.mktemp_project(
            pyproject_toml="""
                [tool.towncrier]
                package = "foobar"
                [tool.towncrier.fragment.feat]
                ignored_field="Bazz"
                [tool.towncrier.fragment.fix]
                [tool.towncrier.fragment.chore]
                name = "Other Tasks"
                showcontent = false
                [tool.towncrier.fragment.auto]
                name = "Automatic"
                check = false
            """
        )
        config = load_config(project_dir)
        expected = {
            "chore": {
                "name": "Other Tasks",
                "showcontent": False,
                "check": True,
            },
            "feat": {
                "name": "Feat",
                "showcontent": True,
                "check": True,
            },
            "fix": {
                "name": "Fix",
                "showcontent": True,
                "check": True,
            },
            "auto": {
                "name": "Automatic",
                "showcontent": True,
                "check": False,
            },
        }
        actual = config.types
        self.assertDictEqual(expected, actual)



from __future__ import annotations

import textwrap

from functools import wraps
from importlib import resources
from pathlib import Path
from subprocess import call
from typing import Any, Callable

from click.testing import CliRunner


def read(filename: str | Path) -> str:
    return Path(filename).read_text()


def write(path: str | Path, contents: str, dedent: bool = False) -> None:
    """
    Create a file with given contents including any missing parent directories
    """
    p = Path(path)
    p.parent.mkdir(parents=True, exist_ok=True)
    if dedent:
        contents = textwrap.dedent(contents)
    p.write_text(contents)


def read_pkg_resource(path: str) -> str:
    """
    Read *path* from the towncrier package.
    """
    return (resources.files("towncrier") / path).read_text("utf-8")


def with_isolated_runner(fn: Callable[..., Any]) -> Callable[..., Any]:
    """
    Run *fn* within an isolated filesystem and add the kwarg *runner* to its
    arguments.
    """

    @wraps(fn)
    def test(*args: Any, **kw: Any) -> Any:
        runner = CliRunner()
        with runner.isolated_filesystem():
            return fn(*args, runner=runner, **kw)

    return test


def setup_simple_project(
    *,
    config: str | None = None,
    extra_config: str = "",
    pyproject_path: str = "pyproject.toml",
    mkdir_newsfragments: bool = True,
) -> None:
    if config is None:
        config = "[tool.towncrier]\n" 'package = "foo"\n' + extra_config
    else:
        config = textwrap.dedent(config)
    Path(pyproject_path).write_text(config)
    Path("foo").mkdir()
    Path("foo/__init__.py").write_text('__version__ = "1.2.3"\n')

    if mkdir_newsfragments:
        Path("foo/newsfragments").mkdir()


def with_project(
    *,
    config: str | None = None,
    pyproject_path: str = "pyproject.toml",
) -> Callable[[Callable[..., Any]], Callable[..., Any]]:
    """Decorator to run a test with an isolated directory containing a simple
    project.

    The files are not managed by git.

    `config` is the content of the config file.
    It will be automatically dedented.

    `pyproject_path` is the path where to store the config file.
    """

    def decorator(fn: Callable[..., Any]) -> Callable[..., Any]:
        @wraps(fn)
        def test(*args: Any, **kw: Any) -> Any:
            runner = CliRunner()
            with runner.isolated_filesystem():
                setup_simple_project(
                    config=config,
                    pyproject_path=pyproject_path,
                )

                return fn(*args, runner=runner, **kw)

        return test

    return decorator


def with_git_project(
    *,
    config: str | None = None,
    pyproject_path: str = "pyproject.toml",
) -> Callable[[Callable[..., Any]], Callable[..., Any]]:
    """Decorator to run a test with an isolated directory containing a simple
    project checked into git.
    Use `config` to tweak the content of the config file.
    Use `pyproject_path` to tweak the location of the config file.
    """

    def decorator(fn: Callable[..., Any]) -> Callable[..., Any]:
        def _commit() -> None:
            call(["git", "add", "."])
            call(["git", "commit", "-m", "Second Commit"])

        @wraps(fn)
        def test(*args: Any, **kw: Any) -> Any:
            runner = CliRunner()
            with runner.isolated_filesystem():
                setup_simple_project(
                    config=config,
                    pyproject_path=pyproject_path,
                )

                call(["git", "init"])
                call(["git", "config", "user.name", "user"])
                call(["git", "config", "user.email", "user@example.com"])
                call(["git", "config", "commit.gpgSign", "false"])
                call(["git", "add", "."])
                call(["git", "commit", "-m", "Initial Commit"])

                return fn(*args, runner=runner, commit=_commit, **kw)

        return test

    return decorator
